# coding=utf-8
"""Musa audio library management tools

Audio file processing libraries

"""
__version__ = '3.7.2'

import os
import sys
import unicodedata

from musa.defaults import MUSA_USER_DIR

class MusaError(Exception):
    pass

def normalized(path, normalization='NFC'):
    """
    Return given path value as normalized unicode string on OS/X,
    on other platform return the original string as unicode
    """
    if sys.platform != 'darwin':
        if not isinstance(path, unicode):
            return unicode(path, 'utf-8')
    if not isinstance(path, unicode):
        path = unicode(path, 'utf-8')
    return unicodedata.normalize(normalization, path)


class CommandPathCache(list):

    """
    Class to represent commands on user's search path.
    """
    def __init__(self):
        list.__init__(self)
        self.paths = None
        self.update()

    def update(self):
        """
        Updates the commands available on user's PATH
        """
        self.paths = []
        self.__delslice__(0, len(self))
        for path in os.getenv('PATH').split(os.pathsep):
            if not self.paths.count(path):
                self.paths.append(path)
        for path in self.paths:
            if not os.path.isdir(path):
                continue
            for cmd in [os.path.join(path, f) for f in os.listdir(path)]:
                if os.path.isdir(cmd) or not os.access(cmd, os.X_OK):
                    continue
                self.append(cmd)

    def versions(self, name):
        """
        Returns all commands with given name on path, ordered by PATH search
        order.
        """
        if not len(self):
            self.update()
        return filter(lambda x: os.path.basename(x) == name, self)

    def which(self, name):
        """
        Return first matching path to command given with name, or None if
        command is not on path
        """
        if not len(self):
            self.update()
        try:
            return filter(lambda x: os.path.basename(x) == name, self)[0]
        except IndexError:
            return None

if not os.path.isdir(MUSA_USER_DIR):
    try:
        os.makedirs(MUSA_USER_DIR)
    except OSError, (ecode, emsg):
        raise MusaError('Error creating directory {0}: {1}'.format(dst_dir, emsg))



# coding=utf-8
"""Musa default settings

Default settings for musa configuration database and commands.

"""

import sys
import os

if sys.platform=='darwin':
    MUSA_USER_DIR = os.path.expanduser('~/Library/Application Support/Musa')
    MUSA_CACHE_DIR = os.path.expanduser('~/Library/Caches/Musa')
else:
    MUSA_USER_DIR = os.path.expanduser('~/.config/musa')
    MUSA_CACHE_DIR = os.path.expanduser('~/.cache/musa')

# Default settings for empty database
INITIAL_SETTINGS = {
    'threads':  4,
    'default_codec': 'mp3',
}



# coding=utf-8
"""Transcoding support

Module for transcoding between file formats

"""

import sys
import os
import shutil
import time
import tempfile

from musa.defaults import MUSA_CACHE_DIR
from musa.cli import ScriptThread, MusaThreadManager
from soundforest.tags import TagError
from soundforest.tree import Tree, Album, Track, TreeError


class TranscoderError(Exception):
    """Exceptions raised by transcoder threads"""

    def __str__(self):
        return self.args[0]


class TranscoderThread(ScriptThread):
    """
    Class to transcode one file from Transcoder queue.
    """

    def __init__(self, index, src, dst, overwrite=False, dry_run=False):
        super(TranscoderThread, self).__init__('convert')
        self.index = index
        self.src = src
        self.dst = dst
        self.overwrite = overwrite
        self.dry_run = dry_run

        if not os.path.isdir(MUSA_CACHE_DIR):
            try:
                os.makedirs(MUSA_CACHE_DIR)
            except OSError, (ecode, emsg):
                raise TranscoderError('Error creating directory {0}: {1}'.format(MUSA_CACHE_DIR, emsg))

    def error(self, message):
        print message
        sys.exit(1)

    def run(self):
        """
        Run the thread, calling source_song.decode() to create a temporary
        wav file and dst_song.encode() to encode this wav file to target file.
        """
        self.status = 'initializing'

        if self.overwrite and os.path.isfile(self.dst.path):
            try:
                if not self.dry_run:
                    os.unlink(self.dst.path)
                else:
                    self.log.debug('remove: {0}'.format(self.dst.path))
            except OSError, (ecode, emsg):
                self.error('Error removing {0}: {1}'.format(self.dst.path, emsg))

        dst_dir = os.path.dirname(self.dst.path)
        if not os.path.isdir(dst_dir):
            try:
                if not self.dry_run:
                    os.makedirs(dst_dir)
            except OSError, (ecode, emsg):
                if os.path.isdir(dst_dir):
                    # Other thread created directory before us, forget it
                    pass
                else:
                    self.error('Error creating directory {0}: {1}'.format(dst_dir, emsg))

        wav = tempfile.NamedTemporaryFile(
            dir=MUSA_CACHE_DIR, prefix='musa-', suffix='.wav'
        )
        src_tmp = tempfile.NamedTemporaryFile(
            dir=MUSA_CACHE_DIR, prefix='musa-', suffix='.{0}'.format(self.src.extension)
        )
        src = Track(src_tmp.name)
        dst_tmp = tempfile.NamedTemporaryFile(
            dir=MUSA_CACHE_DIR, prefix='musa-', suffix='.{0}'.format(self.dst.extension)
        )
        src = Track(src_tmp.name)
        dst = Track(dst_tmp.name)
        if not self.dry_run:
            open(src.path, 'wb').write(open(self.src.path, 'rb').read())

        try:
            decoder = src.get_decoder_command(wav.name)
            encoder = dst.get_encoder_command(wav.name)
        except TreeError, emsg:
            self.error(emsg)


        try:
            if not self.dry_run:
                self.status = 'transcoding'
                self.log.debug('decoding: {0} {1}'.format(self.index, self.src.path))
                self.execute(decoder)
                self.log.debug('encoding: {0} {1}'.format(self.index, self.dst.path))
                self.execute(encoder)
                shutil.copyfile(dst.path, self.dst.path)
            else:
                self.log.debug('decoder: {0}'.format(' '.join(decoder)))
                self.log.debug('encoder: {0}'.format(' '.join(encoder)))
                self.log.debug('target file: {0}'.format(self.dst.path))
            del(wav)

        except TranscoderError, emsg:
            if wav and os.path.isfile(wav):
                try:
                    del(wav)
                except OSError:
                    pass
            self.status = str(e)
            self.error('ERROR transcoding: {0}'.format(emsg))

        if not self.dry_run and not os.path.isfile(self.dst.path):
            self.error('File was not successfully transcoded: {0}'.format(self.dst.path))

        if not self.dry_run:
            try:
                self.status = 'tagging'
                if self.src.tags is not None:
                    self.log.debug('tagging:  {0} {1}'.format(self.index, self.dst.path))
                    if not self.dry_run:
                        if self.dst.tags is not None:
                            if self.dst.tags.update_tags(self.src.tags.as_dict()):
                                self.dst.tags.save()
                        else:
                            # Destination does not support tags
                            pass
            except TagError, emsg:
                self.error(emsg)

        self.status = 'finished'


class MusaTranscoder(MusaThreadManager):
    def __init__(self, threads, overwrite=False, dry_run=False):
        super(MusaTranscoder, self).__init__('convert', int(threads))
        self.overwrite = overwrite
        self.dry_run = dry_run

    def enqueue(self, src, dst):
        if not isinstance(src, Track) or not isinstance(dst, Track):
            raise TranscoderError('Trancode arguments must be track object')

        if not os.path.isfile(src.path):
            raise TranscoderError('No such file: {0}'.format(src.path))

        try:
            src_decoder = src.get_decoder_command('/tmp/test.wav')
        except TreeError, emsg:
            raise TranscoderError(str(emsg))

        try:
            dst_encoder = dst.get_encoder_command('/tmp/test.wav')
        except TreeError, emsg:
            raise TranscoderError(str(emsg))

        self.log.debug('enqueue: {0} -> {1}'.format(src.path, dst.path))
        self.append((src, dst))

    def get_entry_handler(self, index, entry):
        src, dst = entry
        return TranscoderThread(index, src, dst, self.overwrite, self.dry_run)

    def run(self):
        self.log.debug('Transcoding {0} files with {1:d} threads'.format(len(self), self.threads))
        MusaThreadManager.run(self)




# coding=utf-8
"""CLI utilities

Command line utilities for musa

"""

import sys
import os
import time
import logging
import argparse
import tempfile
import signal
import socket
import threading
import subprocess

from soundforest.cli import Script, ScriptCommand, ScriptThread, ScriptThreadManager, ScriptError
from soundforest.prefixes import TreePrefixes
from soundforest.formats import match_metadata, match_codec
from soundforest.tree import Tree, Track, TreeError


class MusaThreadManager(ScriptThreadManager):
    def __init__(self, name, threads=None):
        super(MusaThreadManager, self).__init__(name, threads)

    def enqueue(self, item):
        self.log.debug('enqueue: {0}'.format(src.path))
        self.append((src, dst))

    def run(self):
        if len(self)==0:
            return

        total = len(self)
        while len(self)>0:
            active = threading.active_count()
            if active > self.threads:
                time.sleep(0.5)
                continue

            index = '{0:d}/{1:d}'.format(total-len(self)+1, total)
            t = self.get_entry_handler(index, self.pop(0))
            t.start()

        active = threading.active_count()
        while active > 1:
            time.sleep(0.5)
            active = threading.active_count()


class MusaTagsEditor(ScriptThread):
    def __init__(self, tmpfile):
        super(MusaTagsEditor, self).__init__('musa-edit')
        self.tmpfile = tmpfile

    def run(self):
        self.status = 'edit'

        editor = os.getenv('EDITOR').split()
        if editor is None:
            editor = ['vi']

        cmd = editor + [self.tmpfile]
        p = subprocess.Popen(cmd, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr)
        p.wait()
        self.status = 'finished'


class MusaScript(Script):
    """
    Musa CLI tool setup class
    """

    def edit_tags(self, tags):
        """
        Dump, open and load back a dictionary with EDITOR
        """
        if not isinstance(tags, dict):
            raise ScriptError('Argument not a dictionary')

        tmp = tempfile.NamedTemporaryFile(
            dir=os.getenv('HOME'),
            prefix='.musa-tags',
            suffix='.txt'
        )

        fd = open(tmp.name, 'w')
        for k in sorted(tags.keys()):
            for v in tags[k]:
                fd.write('{0}={1}\n'.format(k, v))
        fd.close()

        editor = MusaTagsEditor(tmp.name)
        editor.start()
        self.wait()

        new_tags = {}
        for l in [l.strip() for l in open(tmp.name, 'r').readlines()]:
            try:
                if l.strip()=='' or l[:1]=='#':
                    continue

                k, v = [x.strip() for x in l.split('=', 1)]
                if k in new_tags.keys():
                    new_tags[k].append(v)
                else:
                    new_tags[k] = [v]

            except ValueError, emsg:
                raise ScriptError('Error parsing new tags from file: {0}'.format(emsg))

        return new_tags

class MusaScriptCommand(ScriptCommand):
    """
    Parent class for musa cli subcommands
    """

    def get_tags(self, track):
        if isinstance(track, Track):
            try:
                return track.tags

            except TreeError, emsg:
                self.log.debug('Error parsing tags from {0}: {1}'.format(track.path, emsg))
                return None

        return None

    def get_codec(self, codec):
        return match_codec(codec)

    def process_tracks(self, trees, tracks, command, **kwargs):
        """
        Execute command for all tracks in trees or track list
        """
        for tree in trees:
            for track in tree:
                command(track=track, **kwargs)

        for track in tracks:
            command(track=track, **kwargs)

    def read_input_to_dict(self, fd):
        tags = {}

        with fd as input:
            for line in [l.rstrip() for l in input]:
                tag, value = None,  None
                for sep in ('=', ' ', '\t'):
                    try:
                        tag, value = [x.strip() for x in line.split(sep, 1)]
                    except ValueError:
                        pass

                if tag is None or value is None:
                    raise ScriptError('Invalid tag input line: {0}'.format(line))
                tags[tag] = unicode(value)

        return tags

    def run(self, args, skip_targets=False):
        """
        Common argument parsing
        """
        args = super(MusaScriptCommand, self).parse_args(args)

        self.prefixes = TreePrefixes()

        if skip_targets:
            return [], [], []

        trees, tracks, metadata = [], [], []
        for path in args.paths:
            if os.path.isdir(path):
                trees.append(Tree(path))

            else:
                try:
                    tracks.append(Track(path))
                except TreeError:
                    match = match_metadata(path)
                    if match is not None:
                        metadata.append(match)

        tracks_found = False
        for d in trees:
            if not len(d):
                continue
            tracks_found = True
            break

        if not tracks_found and not len(tracks) and not len(metadata):
            return [], [], []

        return trees, tracks, metadata





import os
import re
import unittest
from musa import tree

TEST_FILE_PATHS = {
    'load_tests': [
        'foo/bar/baz.m4a',
        'foo/bar/baz.txt',
        'foo/bar/baz.mp3',
        'foo/mp3/flac mp3 aac test file.invalid',
    ],
    'name_tests': [
        '01 Test Name.m4a',
        '01 Test Name.mp3',
        '01 Tracknumber For Text File.txt',
        'No Tracknumber.ogg',
        'Test Name/02 Another Test.flac',
        '01 Test Name/No Tracknumber.flac',
        '01 Test Name/No Tracknumber Text.txt',
        'Test Name/01 Tracknumber for Text File.txt',
    ],
}

class tree_parsing(unittest.TestCase):

    def setUp(self,TEST_ROOT='test/data/tree'):
        self.root = TEST_ROOT
        for root, paths in TEST_FILE_PATHS.items():
            root = os.path.join(self.root, root)

            if not os.path.isdir(root):
                os.makedirs(root)

            for path in [os.path.join(root, x) for x in paths]:
                dir_path = os.path.dirname(path)

                if not os.path.isdir(dir_path):
                    os.makedirs(dir_path)

                if not os.path.isfile(path):
                    open(path,'w').write('\n')

    def tearDown(self):
        for root, paths in TEST_FILE_PATHS.items():
            root = os.path.join(self.root, root)

            for path in [os.path.join(root, x) for x in paths]:
                dir_path = os.path.dirname(path)
                if os.path.isfile(path):
                    os.unlink(path)

            for (root,dirs,files) in os.walk(root,topdown=False):
                os.rmdir(root)

            try:
                os.rmdir(self.root)
            except OSError,(ecode,emsg):
                pass

    def test_tree_items(self):
        t = tree.Tree(
            os.path.join(self.root, 'load_tests')
        )

        self.assertTrue(isinstance(t, tree.Tree))
        self.assertEquals(len(t), 2)

        for track in t:
            self.assertTrue(isinstance(track, tree.Track))
            self.assertTrue(os.path.isfile(track.path))

        for entry in t.files:
            self.assertTrue(isinstance(entry, tuple))
            self.assertEquals(len(entry), 2)
            self.assertTrue(os.path.isfile(os.path.join(entry[0], entry[1])))

    def test_tree_file_count(self):
        # Count music files in tree
        t = tree.Tree(
            os.path.join(self.root, 'load_tests')
        )

        tracks = len(t)
        expected = 2
        self.assertEquals(
            tracks,
            expected,
            'Invalid number of music files in test tree: {0}!={1}'.format(tracks,expected),
        )

        tracks = len(t)
        expected = 2
        self.assertEquals(
            tracks,
            expected,
            'Invalid count after explicit tree reload: {0}!={1}'.format(tracks,expected),
        )

    def test_tree_filter_regexp(self):
        t = tree.Tree(
            os.path.join(self.root, 'name_tests')
        )
        re_test = re.compile('^[0-9]+\s+.*$')

        matches = t.filter_tracks(re_test, re_path=False, re_file=True)
        expected = 3
        self.assertEquals(len(matches), expected)

        for f in matches:
            self.assertTrue(isinstance(f, tuple))
        matches = t.filter_tracks(re_test, re_path=False, re_file=True, as_tracks=True)

        expected = 3
        for f in matches:
            self.assertTrue(isinstance(f, tree.Track))

        expected = 0
        matches = t.filter_tracks(re_test, re_path=True, re_file=False)
        self.assertEquals(len(matches), expected)

        re_test = re.compile('^.*/[0-9]+\s+.*$')
        matches = t.filter_tracks(re_test, re_path=False, re_file=True)
        expected = 0
        self.assertEquals(len(matches), expected)

        expected = 1
        matches = t.filter_tracks(re_test, re_path=True, re_file=False)
        self.assertEquals(len(matches),expected)

suite = unittest.TestLoader().loadTestsFromTestCase(tree_parsing)





import glob
from setuptools import setup, find_packages
from musa import __version__

setup(
    name = 'musa',
    version = __version__,
    license = 'PSF',
    author = 'Ilkka Tuohela',
    author_email = 'hile@iki.fi',
    description = 'Module for music tagging and library management',
    keywords = 'music library tag management',
    url = 'https://github.com/hile/musa',
    package_data = { '': [ '*.md', '*.txt' ] },
    scripts = glob.glob('bin/*'),
    packages = find_packages(),
    install_requires = (
        'soundforest>=3.7.2',
    ),
)





import unittest
from musa import metadata

ARTWORK_PREFIXES = ['albumart','artwork','album','front','back','cover']
ARTWORK_FORMATS = ['jpg','jpeg','png','gif']

VALID_PLAYLIST_NAMES = ['foo.pls','foo.m3u','foo.bar.m3u8']
INVALID_PLAYLIST_NAMES = ['.m3u','foo.mp3','foo.ogg']

class test_metadata(unittest.TestCase):

    def setUp(self):
        self.metadata = metadata.Metadata()

    def tearDown(self):
        del self.metadata

    def test_registration(self):

        class validTestClass(metadata.MetadataFile):
            def __init__(self, path=None):
                super(validTestClass, self).__init__(path, 'Test Class', filenames=['foo'])

        class invalidTestClass(list):
            def __init__(self):
                return

        self.metadata.register_metadata(validTestClass())
        with self.assertRaises(ValueError):
            self.metadata.register_metadata(invalidTestClass())

        self.metadata.unregister_metadata(validTestClass)
        with self.assertRaises(ValueError):
            self.metadata.unregister_metadata(invalidTestClass)

    def test_artwork_files(self):
        """
        Test artwork names for metadata matches
        """
        ARTWORK_NAMES = ['%s.%s'%(name,ext) \
            for name in ARTWORK_PREFIXES \
            for ext in ARTWORK_FORMATS
        ]
        for name in ARTWORK_NAMES:
            m = self.metadata.match(name)
            self.assertIsNotNone(m,'No match for artwork file %s' % name)
            self.assertEquals(m.description,'Album Artwork')

    def test_playlist_files(self):
        for name in VALID_PLAYLIST_NAMES:
            m = self.metadata.match(name)
            self.assertIsNotNone(m,'No match for playlist file %s' % name)
            self.assertEquals(m.description,'Playlist')

        for name in INVALID_PLAYLIST_NAMES:
            m = self.metadata.match(name)
            self.assertIsNone(m,'Invalid name matches metadata: %s' % name)

suite = unittest.TestLoader().loadTestsFromTestCase(test_metadata)





import os
import re
import unittest

from musa.models import Codec
from musa.config import MusaConfigDB
from musa.formats import MusaFileFormat, filter_available_command_list, match_codec


class test_codecs(unittest.TestCase):
    def test_codec_config_keys(self):
        db = MusaConfigDB()

        for name in db.codecs.keys():
            self.assertIsInstance(match_codec(name), Codec)
            self.assertEquals(name, match_codec(name).name)

        for name, codec in db.codecs.items():
            if 'extensions' not in codec.extensions:
                continue

            for ext in codec['extensions']:
                self.assertEquals(name, match_codec(ext).name)

    def test_fileformat_loader(self):
        db = MusaConfigDB()
        for name in db.codecs.keys():
            testpath = os.path.join('/tmp/test.{0}'.format(name))

            musaformat = MusaFileFormat(testpath)
            self.assertEquals(name, musaformat.codec.name)
            self.assertIsInstance(musaformat.get_available_encoders(), list)
            self.assertIsInstance(musaformat.get_available_decoders(), list)

        self.assertEquals(MusaFileFormat('/etc/passwd').codec, None)


suite = unittest.TestLoader().loadTestsFromTestCase(test_codecs)
